/***************************************************************************
 * blitz/alg/fourier.cc     
 * function implementation of vectorized fourier transformation
 *
 *
 * $Id: fourier.cc,v 1.1.1.1 2003/07/09 23:55:52 mdsuresh Exp $
 *
 * Copyright (C) 2002 Suresh kumar Devanathan <mdsuresh@media.mit.edu>
 *
 * Digital Signal Processing Book
 * Copyright (C) 2001 Orfanidis
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Suggestions:          blitz-dev@oonumerics.org
 * Bugs:                 blitz-bugs@oonumerics.org
 *
 * For more information, please see the Blitz++ Home Page:
 *    http://oonumerics.org/blitz/
 *
 ***************************************************************************
 * $Log: fourier.cc,v $
 * Revision 1.1.1.1  2003/07/09 23:55:52  mdsuresh
 *
 *
 * Revision 1.0  2002/10/26 09:22:49  mdsuresh
 * Updated copyright date in headers.
 **/
 
 
#ifndef _VT_FUNCTION_FOURIER_CC
#define _VT_FUNCTION_FOURIER_CC

#ifndef _VT_PACK_H
#error the implementation of fourier transform requires inclusion of <blitz/alg/port.h> \
	prior to the inclusion of this file.
#endif   


namespace _VT_Pack {
	namespace Functions {
		namespace Spectral {
			namespace Fourier {

				int BitReverse(int n, int B )
				{
					int r=0;
					for(; B> 0; B--) {
						r = (r << 1) | ( n & 1)  ; 
						n >>= 1;
					}
					return r;
				}



				template<class T>
					T Shuffle(T elements, int B) 
				{
					T result(_VT_shape(elements));
					_VT_DeclareRangeAll( _VT_RangeAll );
					_VT_PRECHECK( _VT_pow(2,B) ==  _VT_rows(elements) , 
						"The current implementation of FFT tranform "
						"only supports dyalic values.");

					//result = 0;

					for(int i =0; i  < _VT_rows(elements); i++) {
						int  r = BitReverse(i,B);
						Range _r(r), _i(i);
						result(_i) =  elements(_r);                 

					}

					return result;


				}

				template<class T>
					complex<T> conjugate(complex<T> a) 
				{
					return complex<T>(a.real(), -a.imag());
				}
				/*
				template<class T>
				void conjugate(complex<T>& a) {
				a = complex<T>(a.real(), -a.imag());
				}
				*/

				BZ_DECLARE_FUNCTION(conjugate)
					//BZ_DECLARE_FUNCTION_RET(conjugate,void)

					template<class T>
					void DFTMerge(T  elements) 
				{
					double pi = 4. * _VT_atan(1.0);
					int k, i, p, q, M;
					_bz_use(i,k,p,q);
					int N = _VT_rows(elements);
					/*

					_VT_PRECHECK( _VT_pow(2,B) == N , 
						"The current implementation of FFT tranform "
						"only supports dyalic values.");
					*/

					_VT_numtype(T)  V, W;
					M = 2;
					T temp(elements.shape());
					// temp  =0;
					_VT_DeclareRangeAll( _VT_RangeAll );


					while (M <= N) {                      /* two \((M/2)\)-DFTs into one \(M\)-DFT  */
						W =   _VT_exp( _VT_numtype(T)(.0, -2 * pi / M));   /* order-\(M\) twiddle factor */
						V =   _VT_numtype(T)(1., 0.);                  /* successive powers of \(W\) */

						for (k = 0; k < M/2; k++) {          /* index for an \((M/2)\)-DFT */

							_VT_Range p( k + 0, N-1, M);
							_VT_Range q =  p + M/2;

							temp(p) = elements(p) +  V * elements(q);     
							temp(q) = elements(p) -  V * elements(q);

							elements(p) = temp(p);
							elements(q) = temp(q);

							V = V * W;                        /* \(V = VW = W\sp{k+1}\) */
						}
						M = 2 * M;                           /* next stage */

					}

					return;
				}

				template<class T>
					T Transform( T  elements) 
				{
					int B = (int) 
						( _VT_log( double (_VT_rows(elements)) )/ _VT_log(2.0) + .5);
					int N = _VT_rows(elements);

					_VT_PRECHECK( _VT_pow(2,B) == N , 
						"The current implementation of FFT tranform "
						"only supports dyalic values.");

					T temp = Shuffle(elements,B);
					DFTMerge(temp);
					return temp;

				} 

				/*

				#ifdef _VT_ManualVectorize1D         

				// Implement 1D version that calls the 2d-version
				template<class T>
				_VT_Array1D(T) Transform( _VT_Array1D(T)& elements) {
				_VT_Array2D(T) temp( _VT_dataFirst(elements), n, 1);
				return _VT_Array1D(T)( _VT_dataFirst(Transform(temp)) , n ,1);
				} 

				#endif
				*/      


				template<class T>
					T InverseTransform( T  elements) 
				{
					int B = (int) 
						( _VT_log(double( _VT_rows(elements)))/ _VT_log(2.0) + .5);
					int N = _VT_rows(elements);

					_VT_PRECHECK( _VT_pow(2,B) == N , 
						"The current implementation of FFT tranform "
						"only supports dyalic values.");


					T temp = Shuffle( elements ,B);
					temp = conjugate(temp);                      
					DFTMerge(temp);
					temp = conjugate(temp) / ( _VT_numtype(T)(N,0));
					return temp;

				} 

				/*

				#ifdef _VT_ManualVectorize1D         

				// Implement 1D version that calls the 2d-version
				template<class T>
				_VT_Array1D(T) InverseTransform( _VT_Array1D(T)& elements) {
				_VT_Array2D(T) temp( _VT_dataFirst(elements), n, 1);
				return _VT_Array1D(T)( _VT_dataFirst(InverseTransform(temp)) , n ,1);
				} 

				#endif

				*/              

			}
		}
	}
}
#ifdef _VT_MATLAB_INTERFACE

namespace _VT_MATLAB_NAMESPACE {
   using namespace _VT_Pack;
	/* MATLAB-compatible names. */
	/* Export these functions only if the users wants them. */

#ifdef _VT_FOURIER_BLITZ
	template<class T>
		T fft(T input) 
	{
		return _VT_Pack::Functions::Spectral::Fourier::Transform(input);
	}

	template<class T>
		T ifft(T input)
	{
		return _VT_Pack::Functions::Spectral::Fourier::InverseTransform(input);
	}

#endif


}

#endif // _VT_MATLAB_INTERFACE

#endif // _VT_FUNCTION_FOURIER_CC
