
/***************************************************************************
* blitz/alg/cholesky.cc     
* function implementation of vectorized cholesky factorization
*
* $Id: cholesky.cc,v 1.2 2002/10/26 20:55:49 mdsuresh Exp $
*
* Copyright (C) 2002 Suresh kumar Devanathan <mdsuresh@yahoo.com>
* 
* $Id: linalg.h Version 1.1
* Copyright (C) 1998-1999 Brent Worden WordenWare <Brent@Worden.org
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* Suggestions:          blitz-dev@oonumerics.org
* Bugs:                 blitz-bugs@oonumerics.org
*
* For more information, please see the Blitz++ Home Page:
*    http://oonumerics.org/blitz/
*
***************************************************************************
* $Log: cholesky.cc, v $
* Revision 1.0  2002/10/26 20:22:49  mdsuresh
* Updated copyright date in headers.
**/


#ifndef _VT_FUNCTIONS_CHOLESKY_CC
#define _VT_FUNCTIONS_CHOLESKY_CC

#ifndef _VT_PACK_H
#error the implementation of cholesky.cc requires inclusion of <blitz/alg/port.h> \
	prior to the inclusion of this file.
#endif     

namespace  _VT_Pack {
	namespace Functions {
		namespace Matrix {
			namespace Factorization {

				template<class T1, class T2>
					void  CholeskyDecomposition(T1 & a, T2 & p)
					//-------------------------------------------------------------------
					// Given a positive-definite symmetric matrix a, this routine
					// constructs its Colesky decomposition, a = LL'.  On input, only the
					// upper triangle of a need be given; it is not modified.  The
					// Cholesky factor L is returned in the lower traingle of a, except
					// for its diagonal elements which are returned in p.  The routine
					// returns true if the inversion was successful, otherwise it returns
					// false.
					//-------------------------------------------------------------------
				{
					int i, j, k;
					_VT_numtype(T1) _sum;
					int n = _VT_rows(a);
					_VT_PRECHECK( _VT_rows(a) == _VT_cols(a),
						"This operation is only valid on square matrices.\n"
						"Invalid shape:" << _VT_shape(a) );
					_VT_resize(p,n);
					_VT_DeclareRangeAll( _VT_RangeAll );

					p = 0;
					//p(0) = _VT_sqrt(a(0,0));

					//a(_VT_Range(0, n-2), 0) =  a(_VT_Range(1,n-1),0);
					

					for(i = 0; i < n; ++i){
						_sum = a(i,i) - _VT_sum( a( i, _VT_Range(0,i-1)) * a(i, _VT_Range(0,i-1)) ) ;

						_VT_ILLCONDITION( _sum > _VT_numtype(T1)(0),
							"Unable to recover from this ill conditional operation" );

						p(i) = _VT_sqrt(_sum);
						_VT_Range q(i+1,n-1);
						_VT_Range l(0,i-1);
						a(q,i) = (a(i,q) - _VT_sum(a(q,l)(_VT_tensor::i, _VT_tensor::j) * a(i,l)(_VT_tensor::j) , _VT_tensor::j)) / p(i);

					}


				
				}

				template<class T1, class T2, class T3, class T4>
					void CholeskySolve(T1 & a, T2& p, T3& x, T4& b)
					//-------------------------------------------------------------------
					// Solves the set of n linear equations a.x = b, where a is a
					// positive-definite symmetric matrix.  a and p are input as the
					// output of the routine choldcmp.  Only the lower triangle of a is
					// accessed.  b is input as the right-hand side vector.  The solution
					// vector is returned in x[0..n-1].  a, and p are not modified and
					// can be left in place for successive calls with different
					// right-hand sides b.  b is not modified unless you identify b as x
					// in the calling sequence, which is allowed.
					//-------------------------------------------------------------------
				{
					int i, k;
					int n = _VT_rows(a);
					_VT_PRECHECK( _VT_rows(a) == _VT_cols(a),
						"This operation is only valid on square matrices.\n"
						"Invalid shape:" << _VT_shape(a) );
					_VT_resize(x,n);
					_VT_PRECHECK( _VT_rows(a) == _VT_rows(b),
						"Incompatible shape mismatch\n"
						"a:" << _VT_shape(a) << "\n"
						"b:" << _VT_shape(b) << "\n" );
					_VT_PRECHECK( _VT_rows(a) == _VT_rows(p),
						"Incompatible shape mismatch\n"
						"a:" << _VT_shape(a) << "\n"
						"b:" << _VT_shape(p) << "\n" );



					for(i = 0; i < n; i++){
						x(i) = ( b(i) -  _VT_sum(  a(i, _VT_Range(i-1,0,-1) ) * x(_VT_Range(i-1,0,-1) ))) / p(i);
					}
					for(i = n - 1; i >= 0; --i){
						x(i)=  ( x(i) -  _VT_sum( a( _VT_Range(i+1, n-1) , i) * x(_VT_Range(i+1,n-1)))) / p(i);
					}
				}

			} 
		} 
	}
}

#ifdef _VT_MATLAB_INTERFACE

namespace _VT_MATLAB_NAMESPACE {

	template<class T>
		_VT_rhs_< _VT_Array2D( _VT_numtype(T) ) , _VT_Array1D( _VT_numtype(T) ) >
		chol( T & _arr)  
	{
		_VT_Array2D( _VT_numtype(T) ) a = _VT_copy(_arr);
		_VT_Array1D( _VT_numtype(T) ) p;
		_VT_Pack::Functions::Matrix::Factorization::CholeskyDecomposition(a,p);
		return _VT_rhs( a, p);           

	}

	/*template<class T1, class T2, class T3>
	cholsl( T1& a, T2& p, T3& b) {

	}
	*/

}

#endif // _VT_MATLAB_INTERFACE


#endif //  _VT_FUNCTIONS_CHOLESKY_CC
