/* Define RUDEBUG for debugging */
/*************************************************************************
  seqlevelorder.c -- Routines for assigning level numbers for gates and 
                  rearranging gates if they are not in level order 
                  these routines treat flip-flops as having level 1
                  (the same as pi's)
*************************************************************************/
#include "all.h"

/***************************************************************************
  copycircuit(node1, node2) copies contents of node1 to node2
***************************************************************************/
void copycircuit(node1, node2)
     NODE *node1;
     NODE *node2;
{/* copycircuit */

  bcopy(node1, node2, sizeof(NODE));
  outputindex(node2) = 0;

}/* copycircuit */

/****************************************************************************
     s_enqueuefanouts() enqueues fanouts connected to a gate
     Inputs: gatenmbr --> gate number whose fanouts are to be queued.
     Function: Queue events to propagate the signal through all fanout gates
               connected to this gate.
*****************************************************************************/
void s_enqueuefanouts(gate)
     NODE *gate;
{  /* s_enqueuefanouts */
  int cfindex          /* Gate input index -- number I/P's - 1 */;

  for(cfindex = 0;cfindex < nmbroutputs (gate); ++cfindex)
    {
      s_enqueue(fanoutgate(gate, cfindex));
    }
}  /* s_enqueuefanouts */

/************************************************************************
  Routine to assign new gate numbers 
  Loops through all the PI's, gates and PO's and assigns gate #'s 
  Assumes that the PI's, gates and PO's are already in levelorder.
************************************************************************/
void assign_gate_numbers(ckt)
     CIRCUIT *ckt;
{ /* assign_gate_numbers */
  int i;
  int gate_counter=1;

  for (i=0; i < nmbr_pi(ckt); i++)
    {
      gate_number(PIs(ckt, i)) = gate_counter++;
    }
  for (i=0; i < nmbr_gates(ckt); i++)
    {
      gate_number(gatesx(ckt, i)) = gate_counter++;
    }
  for (i=0; i < nmbr_po(ckt); i++)
    {
      gate_number(POs(ckt, i)) = gate_counter++;
    }
} /* assign_gate_numbers */

/* Compute sequential level number for gate */
void comp_level (ckt, gate)
     CIRCUIT *ckt;
     NODE * gate;
{ /* comp_level */
  int j;
  NODE * inputdriver;
  NODE * fanoutx;
  unsigned int maxinplevel;
  unsigned int newlevel;

  /* For each output from this gate */
/*  for (i = 0; i < nmbroutputs (gate); i++) */
/*    {*/
     /**********************************************************************
             For debugging, we print gate pointer values */
#ifdef RUDEBUG
  fprintf(resultsout, "gate %d\n", gate_number(gate));
#endif
/*     ***********************************************************************/

  maxinplevel = 0;
  /* Get the maximum gate level of each input to this gate */
  /* Avoiding gates with undefined levels */
  for (j = 0; j < nmbrinputs (gate); j++)
    {
      inputdriver = faningate (gate, j);
      if (levelnumber (inputdriver) > maxinplevel)
	maxinplevel = levelnumber (inputdriver);
    }
  /* Assign this gate max input level number + 1 */
  if (maxinplevel > 0)
    {
      if (gatefunction (gate) == p_po)
	newlevel = maxinplevel;
      else
	newlevel = maxinplevel + 1;
    }
  else
    newlevel = maxinplevel;
  /* If the level number of this gate changed */
  if ((newlevel != levelnumber (gate)) || (gatefunction (gate) == p_pi))
    {
      if (newlevel > 0)
	{
	  levelnumber (gate) = newlevel;
#ifdef RUDEBUG
	  fprintf(resultsout, 
		  "gate %d -- level %d type ", gate_number(gate), levelnumber (gate));
	  switch (gatefunction (gate))
	    {
	    case p_pi: fprintf(resultsout, "pi\n");
	      break;
	    case p_po: fprintf(resultsout, "po\n");
	      break;
	    case p_and: fprintf(resultsout, "and\n");
	      break;
	    case p_nand: fprintf(resultsout, "nand\n");
	      break;
	    case p_or: fprintf(resultsout, "or\n");
	      break;
	    case p_nor: fprintf(resultsout, "nor\n");
	      break;
	    case p_xor: fprintf(resultsout, "xr\n");
	      break;
	    case p_cxor: fprintf(resultsout, "xor\n");
	      break;
	    case p_equiv: fprintf(resultsout, "equivalence\n");
	      break;
	    case p_d_ff: fprintf(resultsout, "dff\n");
	      break;
	    case p_buffer: fprintf(resultsout, "buffer\n");
	      break;
	    case p_not: fprintf(resultsout, "inverter\n");
	      break;
	    case p_f_out: fprintf(resultsout, "fanout\n");
	      break;
	    case p_unknown: fprintf(resultsout, "unknown\n");
	      break;
	    }
#endif
	}
	  
      /* Call comp_level on each gate fanout */
      for (j = 0; j < nmbroutputs (gate); j++)
	{
	  fanoutx = fanoutgate (gate, j);
	  /* Do not queue for flip-flops */
	  if (gatefunction (fanoutx) != p_d_ff)
	    comp_level (ckt, fanoutx);
	}
    }
/*    }*/
  return;
} /* comp_level */

/************************************************************************
  assignlevelnumber() assigns the correct level numbers to the gates.
*************************************************************************/
void assignlevelnumber (circuit)
     CIRCUIT *circuit;
{ /* assignlevelnumber */
  int i, j;
  NODE *igate;
  int gatelevel;

  /* Initialization */
  for (i = 0; i < nmbr_pi (circuit); i++)
    {
      igate = PIs (circuit, i);
      levelnumber (igate) = 1;
#ifdef RUDEBUG
       fprintf(resultsout,
	       "gate %d -- level %d type pi\n", gate_number(igate), levelnumber (igate));
#endif
    }

  for (i = 0; i < (nmbr_po (circuit)); i++)
    {
      igate = POs (circuit, i);
      levelnumber (igate) = 0;
#ifdef RUDEBUG
       fprintf(resultsout, 
	       "gate %d -- level %d type po\n", gate_number(igate), levelnumber (igate));
#endif
    }

  for (i = 0; i < (nmbr_gates (circuit)); i ++)
    {
      igate = gatesx (circuit, i);
      levelnumber (igate) = 0;
#ifdef RUDEBUG
      fprintf(resultsout,
	       "gate %d -- level %d type ", gate_number(igate), levelnumber (igate));
      switch (gatefunction (igate))
	{
	case p_pi: fprintf(resultsout, "pi\n");
	  break;
	case p_po: fprintf(resultsout, "po\n");
	  break;
	case p_and: fprintf(resultsout, "and\n");
	  break;
	case p_nand: fprintf(resultsout, "nand\n");
	  break;
	case p_or: fprintf(resultsout, "or\n");
	  break;
	case p_nor: fprintf(resultsout, "nor\n");
	  break;
	case p_xor: fprintf(resultsout, "xr\n");
	  break;
	case p_cxor: fprintf(resultsout, "xor\n");
	  break;
	case p_equiv: fprintf(resultsout, "equivalence\n");
	  break;
	case p_d_ff: fprintf(resultsout, "dff\n");
	  break;
	case p_buffer: fprintf(resultsout, "buffer\n");
	  break;
	case p_not: fprintf(resultsout, "inverter\n");
	  break;
	case p_f_out: fprintf(resultsout, "fanout\n");
	  break;
	case p_unknown: fprintf(resultsout, "unknown\n");
	  break;
	}
#endif
    }

  for (i = 0; i < (nmbr_pi (circuit)); i ++)
    {
      igate = PIs (circuit, i);
      comp_level (circuit, igate);
    }

  for (i = 0; i < (nmbr_gates (circuit)); i ++)

    {
      igate = gatesx (circuit, i);
      /* If this is a D flip-flop, set its level to 1 and update its fanout
	 gate levels */
      if (gatefunction (igate) == p_d_ff)
	{
	  levelnumber (igate) = 1;
#ifdef RUDEBUG
          fprintf(resultsout, "gate %d -- level %d dff\n", gate_number(igate),
		 levelnumber (igate));
#endif
	  for (j = 0; j < nmbroutputs (igate); ++ j)
	    {
	      comp_level (circuit, fanoutgate (igate, j));
	    }
	}
    }

  /***************************************************************************
   The above standard  level number routine assigns the level numbers to gates 
   and leaves primary outputs and pseudo primary outputs. So we will calcuate
   the level numbers of primary outputs and pseudo primary outputs, by adding
   1 to the level number of their fanin gate. Seq_PO list is the list of POs
   followed by PPOs (Pseudo Primary Outputs).
  ***************************************************************************/

  /* For debugging print the level numbers of each gate  */
#ifdef RUDEBUG
     for(i = 0; i < nmbr_pi(circuit); i++)
        fprintf(resultsout, "gate %d:  level %d type pi\n",gate_number(PIs(circuit,i)),
	       levelnumber(PIs(circuit,i)));
     for(i = 0; i < nmbr_gates(circuit); i++)
       {
        fprintf(resultsout, "gate %d:  level %d type ",gate_number(gatesx(circuit,i)),
	       levelnumber(gatesx(circuit,i)));
	switch (gatefunction (gatesx (circuit, i)))
	  {
	  case p_pi: fprintf(resultsout, "pi\n");
	    break;
	  case p_po: fprintf(resultsout, "po\n");
	    break;
	  case p_and: fprintf(resultsout, "and\n");
	    break;
	  case p_nand: fprintf(resultsout, "nand\n");
	    break;
	  case p_or: fprintf(resultsout, "or\n");
	    break;
	  case p_nor: fprintf(resultsout, "nor\n");
	    break;
	  case p_xor: fprintf(resultsout, "xr\n");
	    break;
	  case p_cxor: fprintf(resultsout, "xor\n");
	    break;
	  case p_equiv: fprintf(resultsout, "equivalence\n");
	    break;
	  case p_d_ff: fprintf(resultsout, "dff\n");
	    break;
	  case p_buffer: fprintf(resultsout, "buffer\n");
	    break;
	  case p_not: fprintf(resultsout, "inverter\n");
	    break;
	  case p_f_out: fprintf(resultsout, "fanout\n");
	    break;
	  case p_unknown: fprintf(resultsout, "unknown\n");
	    break;
	  }
      }	
     for(i = 0; i < nmbr_po(circuit); i++)
        fprintf(resultsout, "gate %d:  level %d type po\n",gate_number(POs(circuit,i)),
	       levelnumber(POs(circuit,i)));
#endif

  circuit->highest_level = 0;
  for (i = 0; i < nmbr_gates (circuit);  i++)
    {
      gatelevel = levelnumber (gatesx (circuit, i));
      if (gatelevel == 0)
	{
	  fprintf (resultsout,
		  "assignlevelnumber: level # not assigned -- gate number %d.\n",
		  i);
	  fatal_error ("assignlevelnumber", COMPUTATION_ERROR, NULL);
	}
      else if (gatelevel > circuit->highest_level)
	circuit->highest_level = gatelevel;
    }

#ifdef RUDEBUG
  fprintf(resultsout, "highest_level = %d.\n", circuit->highest_level);
#endif
  return;

} /* assignlevelnumber */

/*****************************************************************************
  Sort the gate list by level order
*****************************************************************************/
void sort_by_level(node_list, l, r)
     NODE *node_list[];
     int l, r;
{ /* sort_by_level */
  int i, j, k, m;
  NODE **tmp_node_list;

  tmp_node_list = (NODE **)calloc(r+1, sizeof(NODE *));

  if (r > l)
    {
      m = (r+l)/2;
      sort_by_level(node_list, l, m);
      sort_by_level(node_list, m+1, r);

      for (i = m+1; i > l; i--)
	tmp_node_list[i-1] = node_list[i-1];
      for (j = m; j < r; j++)
	tmp_node_list[r+m-j] = node_list[j+1];
      for (k = l; k <= r; k++)
	node_list[k] =
	  (levelnumber(tmp_node_list[i]) < levelnumber(tmp_node_list[j])) ?
	    tmp_node_list[i++] : tmp_node_list[j--];
    }

  free(tmp_node_list);
} /* sort_by_level */

/* update indicies of gates */
void update_gate_index(ckt)
     CIRCUIT *ckt;
{ /* update_gate_index */
  int i;
  
  for (i=0; i<nmbr_pi(ckt); i++)
    {
      gate_index(PIs(ckt, i)) = i;
    }

  for (i=0; i < nmbr_gates(ckt); i++)
    {
      gate_index(gatesx(ckt, i)) = i;
    }

  for (i=0; i < nmbr_po(ckt); i++)
    {
      gate_index(POs(ckt, i)) = i;
    }
} /* update_gate_index */

void arrange_in_levelorder(ckt)
     CIRCUIT *ckt;
{ /* arrange_in_levelorder */
  assignlevelnumber(ckt);
  sort_by_level(gate_list(ckt), 0, nmbr_gates(ckt)-1);
/*  assign_gate_numbers(ckt); */
  update_gate_index(ckt);
} /* arrange_in_levelorder */

